
#include "interfaces/SMPDBInterface.h"
#include "interfaces/abstract/all.h"
#include "interfaces/idapro/all.h"


static bool IsChunkUnshared(ea_t ChunkAddr, ea_t FuncHeadStart, ea_t FuncHeadEnd) 
{
        bool Unshared = true;
        SMP_xref_t CurrXrefs;

        // See if any xref reaches this ChunkAddr from outside the FuncHead range.
        for (bool ok = CurrXrefs.SMP_first_to(ChunkAddr, XREF_ALL);     ok;     ok = CurrXrefs.SMP_next_to()) {
                ea_t FromAddr = CurrXrefs.GetFrom();
                if ((FromAddr != 0) && (CurrXrefs.GetIscode())) {
                        // We found a code xref that comes to the ChunkAddr. Whether it is a fall-through or
                        //  a jump/call, it is unshared only if it comes from within the FuncHead address range.
                        if ((FromAddr < FuncHeadStart) || (FromAddr >= FuncHeadEnd)) {
                                Unshared = false;
                                break;
                        }
                }
        }

        return Unshared;
} // end of SMPProgram::IsChunkUnshared()



void STARS_IDA_Function_t::MarkSharedChunks()
{

	char name[1000];
	ea_t FirstEA=this->get_startEA();
	::get_func_name(FirstEA,name,sizeof(name));
        // Determine if we are dealing with shared chunks.
        ea_t FuncHeadLastAddr = 0;
        size_t ChunkCounter = 0;
        //func_tail_iterator_t FuncTail((func_t*)*dynamic_cast<STARS_IDA_Function_t*>(this->GetFuncInfo()));
        func_tail_iterator_t FuncTail(the_func);// (func_t*)*(dynamic_cast<STARS_IDA_Function_t*>(this->GetFuncInfo())));
        for (bool ChunkOK = FuncTail.main(); ChunkOK; ChunkOK = FuncTail.next()) 
	{
                const area_t &CurrChunk = FuncTail.chunk();
                ++ChunkCounter;
                if (1 == ChunkCounter) { // head chunk
                        FuncHeadLastAddr = CurrChunk.endEA;
                }
                else { // a tail chunk
                        if (IsChunkUnshared(CurrChunk.startEA, FirstEA, FuncHeadLastAddr)) 
			{
//                                this->UnsharedChunks = true;
//                                SMP_msg("INFO: Interface Found unshared tail chunk for %s at %lx\n",
//                                        name, (unsigned long) CurrChunk.startEA);
                        }
                        else 
			{
                                this->SharedChunks = true;
//                                SMP_msg("INFO: Interface Found tail chunk for %s at %lx\n",
//                                        name, (unsigned long) CurrChunk.startEA);
                        }
                }
        }

}

bool STARS_IDA_Function_t::IsMultiEntry()
{
	ea_t FirstEA=this->get_startEA();
        func_tail_iterator_t FuncTail(the_func);// (func_t*)*(dynamic_cast<STARS_IDA_Function_t*>(this->GetFuncInfo())));
        size_t CallTargetCount = 0; // how many addresses in this function are called?

        if (this->HasSharedChunks() || this->UnsharedChunks) {
                for (bool ChunkOK = FuncTail.main(); ChunkOK; ChunkOK = FuncTail.next()) {
                        const area_t &CurrChunk = FuncTail.chunk();
                        ea_t addr = CurrChunk.startEA; // Start with just the beginning addrs of chunks.
                        // Determine whether the instruction is a call target by looking
                        //  at its cross references and seeing if it has "TO" code xrefs.
                        SMP_xref_t xrefs, Distant_xrefs;
                        for (bool ok = xrefs.SMP_first_to(addr, XREF_FAR); ok; ok = xrefs.SMP_next_to()) {
                                ea_t DistantAddr = xrefs.GetFrom();
                                if ((DistantAddr != 0) && (xrefs.GetIscode())) {
                                        // Now we see if the distant instruction is in another function.
                                        STARS_Function_t *SourceFunc = SMP_get_func(DistantAddr);
                                        if (NULL != SourceFunc) {
                                                if (SourceFunc->get_startEA() != FirstEA) {
                                                        ++CallTargetCount;
                                                        break;
                                                }
                                        }
                                }
                        } // end for all xrefs
                } // end for all chunks
        }

        return  CallTargetCount > 1;
} 

